//==============================================================================
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//==============================================================================

//==============================================================================
// File: Main.cpp
// Author: Jarmo Hekkanen <jarski@2ndpoint.fi>
//------------------------------------------------------------------------------
// Includes
#include <iostream>
#include <stdexcept>
#include <string>
#include <list>
#include <fstream>
#include <SDL.h>
#include <SDL_image.h>
#include <stdlib.h>
#include <time.h>
#include "cParticleSystem.hpp"
#ifdef SOUND
	#include <SDL_mixer.h>
#endif
//==============================================================================

// Needed for gcc > 3.0, thanks Bert Salaets
using namespace std;

SDL_Surface *pScreen = NULL;
SDL_Surface *pBackground = NULL;
SDL_Surface *pBitmaps = NULL;
SDL_Rect GameOverRect = {0, 0, 150, 100};
SDL_Rect CrosshairRect = {129, 100, 20, 20};
SDL_Rect TextRect = {0, 100, 100, 20};
SDL_Rect TankRect = {0, 120, 68, 20};
SDL_Rect MarineRect = {68, 120, 10, 14};
SDL_Rect ExplosionDecalRect = {78, 120, 24, 9};
SDL_Rect SandDecalRect = {102, 120, 4, 9};
SDL_Rect BloodDecalRect = {106, 120, 8, 9};

#ifdef SOUND
Mix_Chunk *pGunSound = NULL;
int GunChannel = 1;
Mix_Chunk *pRicochetSound = NULL;
Mix_Chunk *pExplosionSound = NULL;
Mix_Chunk *pDeathSound = NULL;

Mix_Chunk *LoadSound (string filename);
#endif

SDL_Surface *LoadImage (string filename, bool colorkey=false);
void CleanUp (void);
SDL_Rect MakeRect (Sint16 x, Sint16 y, Uint16 w, Uint16 h);
SDL_Color MakeColor (Uint8 r, Uint8 g, Uint8 b);
void SpawnUnits (void);
void Print (int x, int y, int num);
void CheckPosition (int x, int y);
void GameOver (void);
void MakeBlood (Uint16 x, Uint16 y);
void MakeExplosion (Uint16 x, Uint16 y);
void MakeSand (Uint16 x, Uint16 y);
void RenderCrosshair (void);
void BloodDecal (Uint16 x, Uint16 y);
void ExplosionDecal (Uint16 x, Uint16 y);
void SandDecal (Uint16 x, Uint16 y);
void LoadHighscores (void);
void SaveHighscores (void);

struct UnitData
{
	UnitData (void): hits(0) {};
	bool Hit (int x, int y, int w, int h);
	
	int x;
	int y;
	int hits;
};


list<UnitData> Tanks;
list<UnitData> Marines;
list<cParticleSystem> ParticleSystems;

Uint16 Kills = 0;
Uint16 Score = 0;

Uint16 HighscoreKills = 0;
Uint16 HighscoreScore = 0;

bool Running = true;

//==============================================================================
// Main function
//------------------------------------------------------------------------------
int main (int argc, char *argv[])
{
	try
	{
		// Init SDL
		if ( SDL_Init (SDL_INIT_VIDEO) != 0 )
		{
			cerr << "Unable to init SDL: " << SDL_GetError () << endl;
			return 1;
		}
		atexit (CleanUp);
		
		// Set caption
		SDL_WM_SetCaption ("OilWar", NULL);
		
		// Set video mode
		pScreen = SDL_SetVideoMode (640, 480, 16, SDL_HWSURFACE|SDL_DOUBLEBUF);
		if ( pScreen == NULL )
		{
			cerr << "Unable to set video mode: " << SDL_GetError () << endl;
			return 1;
		}
		
		#ifdef SOUND
		// Open audio
		if ( Mix_OpenAudio (22050, AUDIO_S16, 2, 4096) )
		{
			cerr << "Unable to open audio: " << Mix_GetError () << endl;
			return 1;
		}
		#endif
		
		// Random seed
		srand (time (NULL));
		
		// Load images
		pBackground = LoadImage ("background.png");
		pBitmaps = LoadImage ("bitmaps.png", true);
		
		#ifdef SOUND
		// Load sounds
		pGunSound = LoadSound ("singlegunshot.wav");
		pRicochetSound = LoadSound ("bulletricochet.wav");
		pExplosionSound = LoadSound ("explode.wav");
		pDeathSound = LoadSound ("gasp.wav");
		#endif
		
		// Load highscores
		LoadHighscores ();
		
		// Hide cursor
		SDL_ShowCursor (false);

		// Blit background image
		//SDL_BlitSurface (pBackground, NULL, pScreen, NULL);
		SDL_Flip (pScreen);
		
		// Timing
		Uint32 now, deltaTime, lastTime = SDL_GetTicks (), lastMove = SDL_GetTicks ();
		
		list<UnitData>::iterator marine;
		list<UnitData>::iterator tank;
		list<cParticleSystem>::iterator system, tempSystem;
		
		SpawnUnits ();
		
		// Main loop
		SDL_Event event;
		SDL_Rect srcRect, dstRect;
		bool hit = false;
		while ( Running )
		{
			// Poll the event queue
			while ( SDL_PollEvent (&event) != 0 )
			{
				switch ( event.type )
				{
					case SDL_QUIT:
						Running = false;
						break;
					case SDL_KEYDOWN:
						switch ( event.key.keysym.sym )
						{
							case SDLK_ESCAPE:
								Running = false;
								break;
							case SDLK_F10:
								cout << "Saving screenshot to shot.bmp" << endl;
								SDL_SaveBMP (pScreen, "shot.bmp");
								break;
							case SDLK_F11:
								SDL_WM_ToggleFullScreen (pScreen);
								break;
							default:
								break;
						}
					case SDL_MOUSEBUTTONDOWN:
						hit = false;
						marine = Marines.begin ();
						while ( marine != Marines.end () )
						{
							if ( marine->Hit (event.button.x, event.button.y, MarineRect.w, MarineRect.h) )
							{
								BloodDecal (marine->x, marine->y);
								Marines.erase (marine);
								Kills++;
								Score += 10;
								SpawnUnits ();
								MakeBlood (event.button.x, event.button.y);
								hit = true;		
								
								#ifdef SOUND
								// Sounds
								Mix_PlayChannel (-1, pDeathSound, 0);
								Mix_PlayChannel (GunChannel, pGunSound, 0);
								#endif
								
								break;
							}
							marine++;
						}
						tank = Tanks.begin ();
						while ( tank != Tanks.end () )
						{
							if ( tank->Hit (event.button.x, event.button.y, TankRect.w / 2, TankRect.h) )
							{
								#ifdef SOUND
								// Sounds
								Mix_PlayChannel (GunChannel, pGunSound, 0);
								#endif

								tank->hits++;
								hit = true;
								if ( tank->hits >= 5 )
								{
									ExplosionDecal (tank->x, tank->y);
									Tanks.erase (tank);
									Kills++;
									Score += 100;
									SpawnUnits ();
									MakeExplosion (event.button.x, event.button.y);
									
									#ifdef SOUND
									// Sounds
									Mix_PlayChannel (-1, pExplosionSound, 0);
									#endif

									break;
								}
								#ifdef SOUND
								else
									Mix_PlayChannel (-1, pRicochetSound, 0);
								#endif
							}
							tank++;
						}
						if ( !hit )
						{
							if ( event.button.y > 250 && event.button.x <= 250 + int (float (event.button.y - 250) / 200.0f * 180.0f) )								
							{
								MakeSand (event.button.x, event.button.y);
								SandDecal (event.button.x, event.button.y);
								
								#ifdef SOUND
								// Sounds
								Mix_PlayChannel (GunChannel, pGunSound, 0);
								#endif
							}
						}
						break;
					default:
						break;
				}
			}
						
			// Calculate delta time
			now = SDL_GetTicks ();
			deltaTime = now - lastTime;
			lastTime = now;
			
			// Blit background image
			SDL_BlitSurface (pBackground, NULL, pScreen, NULL);
			
			// Render tanks
			tank = Tanks.begin ();
			while ( tank != Tanks.end () )
			{
				if ( tank->x % 2 == 0 )
					srcRect = MakeRect (TankRect.x, TankRect.y, TankRect.w / 2, TankRect.h);
				else
					srcRect = MakeRect (TankRect.x +  TankRect.w / 2, TankRect.y, TankRect.w / 2, TankRect.h);
				dstRect = MakeRect (tank->x, tank->y, 0, 0);
				
				SDL_BlitSurface (pBitmaps, &srcRect, pScreen, &dstRect);
				tank++;
			}
			
			// Render marines
			marine = Marines.begin ();
			while ( marine != Marines.end () )
			{
				dstRect = MakeRect (marine->x, marine->y, 0, 0);
				if ( marine->x % 2 == 0 )
					dstRect.y--;
				SDL_BlitSurface (pBitmaps, &MarineRect, pScreen, &dstRect);
				marine++;
			}
			
			// Render particle systems
			system = ParticleSystems.begin ();
			while ( system != ParticleSystems.end () )
			{
				system->Render (pScreen);
				system++;
			}
			
			// Print kills & score
			Print (490, 40, Kills);
			Print (490, 100, Score);
			
			// Print highscores
			Print (20, 40, HighscoreKills);
			Print (20, 100, HighscoreScore);
			
			// Render crosshair
			RenderCrosshair ();
			
			
			// Update screen
			SDL_Flip (pScreen);
			
			if ( lastMove + 100 < now )
			{
				list<UnitData>::iterator tank = Tanks.begin ();
				while ( tank != Tanks.end () )
				{
					tank->x++;
					CheckPosition (tank->x, tank->y);
					tank++;
				}
				list<UnitData>::iterator marine = Marines.begin ();
				while ( marine != Marines.end () )
				{
					marine->x++;
					CheckPosition (marine->x, marine->y);
					marine++;
				}

				lastMove = now;
			}
			
			// Update particle systems
			system = ParticleSystems.begin ();
			while ( system != ParticleSystems.end () )
			{
				if ( !system->Update (deltaTime) )
				{
					tempSystem = system;
					system++;
					ParticleSystems.erase (tempSystem);
				}
				else
					system++;
			}
		}
		
		return 0;
	}
	catch ( exception &x )
	{
		cerr << "Exception caught: " << x.what () << endl;
	}
	catch ( ... )
	{
		cerr << "Unknown exception caught" << endl;
	}
	return 1;
}
//==============================================================================

//==============================================================================
// Load image from file
//------------------------------------------------------------------------------
SDL_Surface *LoadImage (string filename, bool colorkey)
{
	string completeFilename = DATA_DIR;
	completeFilename.append (filename);
	
	SDL_Surface *pTemp = IMG_Load (completeFilename.c_str ());
	if ( pTemp == NULL )
	{
		pTemp = IMG_Load (filename.c_str ());
		if ( pTemp == NULL )
		{
			cerr << "Unable to load image " << completeFilename.c_str () << " or " << filename.c_str () << endl;
			throw runtime_error ("Unable to load resources");
		}
	}
	if ( colorkey )
		SDL_SetColorKey (pTemp, SDL_SRCCOLORKEY|SDL_RLEACCEL, SDL_MapRGB (pTemp->format, 255, 0, 255));
	
	SDL_Surface *pTemp2 = SDL_DisplayFormat (pTemp);
	SDL_FreeSurface (pTemp);
	if ( pTemp2 == NULL )
	{
		cerr << "Unable to convert image: " << SDL_GetError () << endl;
		throw runtime_error ("Unable to load resources");
	}
	
	return pTemp2;
}
//==============================================================================

//==============================================================================
// Free data etc.
//------------------------------------------------------------------------------
void CleanUp (void)
{
	if ( Kills > HighscoreKills || Score > HighscoreScore )
		SaveHighscores ();
	
	// Free images
	SDL_FreeSurface (pBackground);
	SDL_FreeSurface (pBitmaps);
	
	#ifdef SOUND
	// Free sounds
	Mix_FreeChunk (pGunSound);
	Mix_FreeChunk (pExplosionSound);
	Mix_FreeChunk (pDeathSound);
		
	// Close audio
	Mix_CloseAudio ();
	#endif
	
	// Show cursor
	SDL_ShowCursor (true);
	
	// Quit SDL
	SDL_Quit ();
}
//==============================================================================

//==============================================================================
// Build SDL_Rect structure
//------------------------------------------------------------------------------
SDL_Rect MakeRect (Sint16 x, Sint16 y, Uint16 w, Uint16 h)
{
	SDL_Rect temp = {x, y, w, h};
	return temp;
}
//==============================================================================

//==============================================================================
// Build SDL_Color structure
//------------------------------------------------------------------------------
SDL_Color MakeColor (Uint8 r, Uint8 g, Uint8 b)
{
	SDL_Color temp = {r, g, b};
	return temp;
}
//==============================================================================

//==============================================================================
// Spawn units
//------------------------------------------------------------------------------
void SpawnUnits (void)
{
	if ( Kills % 5 == 0 )
	{
		UnitData tank;
		tank.x = -TankRect.w / 2;
		tank.y = int (310.0f + 120.0f * rand () / (RAND_MAX + 1.0f));
		Tanks.push_back (tank);
	}
	UnitData marine;
	marine.x = -10;
	marine.y = int (310.0f + 120.0f * rand () / (RAND_MAX + 1.0f));;
	Marines.push_back (marine);
	if ( Kills % 3 == 0 )
	{
		marine.y = int (310.0f + 120.0f * rand () / (RAND_MAX + 1.0f));;
		Marines.push_back (marine);
	}
}
//==============================================================================

//==============================================================================
// Is the unit hit
//------------------------------------------------------------------------------
bool 
UnitData::Hit (int mouseX, int mouseY, int w, int h)
{
	if ( mouseX >= x && mouseX < x + w && mouseY >= y && mouseY < y + h )
		return true;
	return false;
}
//==============================================================================

//==============================================================================
// Print number
//------------------------------------------------------------------------------
void Print (int x, int y, int num)
{
	char buffer[10];
	snprintf (buffer, 10, "%09i", num);

	SDL_Rect srcRect, dstRect;
	
	srcRect = MakeRect (TextRect.x, TextRect.y, TextRect.w / 10, TextRect.h);
	dstRect = MakeRect (x, y, 0, 0);
	for ( int i = 0; i < 10; i++ )
	{
		dstRect.x += srcRect.w;
		if ( isdigit (buffer[i]) )
		{
			srcRect.x = TextRect.x + (buffer[i] - 48) * srcRect.w;
			SDL_BlitSurface (pBitmaps, &srcRect, pScreen, &dstRect);
		}
	}
}
//==============================================================================

//==============================================================================
// Check enemy position
//------------------------------------------------------------------------------
void CheckPosition (int x, int y)
{
	y -= 310;
	if ( x >= 290 + int (float (y) / 120.0f * 100.0f) )
		GameOver ();
}
//==============================================================================

//==============================================================================
// Show game over
//------------------------------------------------------------------------------
void GameOver (void)
{
	SDL_Rect dstRect;
	
	dstRect = MakeRect ((pScreen->w - GameOverRect.w) / 2, (pScreen->h - GameOverRect.h) / 2, 0, 0);
	SDL_BlitSurface (pBitmaps, &GameOverRect, pScreen, &dstRect);
	SDL_UpdateRects (pScreen, 1, &dstRect);
	
	// Show cursor
	SDL_ShowCursor (true);
	
	SDL_Event event;
	while ( Running )
	{
		SDL_WaitEvent (&event);
		switch ( event.type )
		{
			case SDL_QUIT:
				Running = false;
				break;
			case SDL_KEYDOWN:
				if ( event.key.keysym.sym == SDLK_ESCAPE )
					Running = false;
				break;
			default:
				break;
		}
	}
}
//==============================================================================

//==============================================================================
// Spawn blood particle system
//------------------------------------------------------------------------------
void MakeBlood (Uint16 x, Uint16 y)
{
	ParticleSystems.push_back (cParticleSystem (x, y, 15,
			MakeColor (255, 0, 0), MakeColor (50, 0, 0),
			2, 10));
	ParticleSystems.push_back (cParticleSystem (x, y, 50,
			MakeColor (200, 0, 0), MakeColor (20, 0, 0),
			1, 20));
}
//==============================================================================

//==============================================================================
// Spawn explosion particle system
//------------------------------------------------------------------------------
void MakeExplosion (Uint16 x, Uint16 y)
{
	ParticleSystems.push_back (cParticleSystem (x, y, 40,
			MakeColor (232, 56, 0), MakeColor (50, 0, 0),
			2, 25));
	ParticleSystems.push_back (cParticleSystem (x, y, 40,
			MakeColor (225, 235, 0), MakeColor (23, 54, 0),
			2, 25));
	ParticleSystems.push_back (cParticleSystem (x, y, 40,
			MakeColor (255, 0, 0), MakeColor (150, 0, 0),
			1, 35));
}
//==============================================================================

//==============================================================================
// Spawn sand particle system
//------------------------------------------------------------------------------
void MakeSand (Uint16 x, Uint16 y)
{
	ParticleSystems.push_back (cParticleSystem (x, y, 20,
			MakeColor (160, 160, 110), MakeColor (120, 110, 60),
			1, 15));
}
//==============================================================================

//==============================================================================
// Render crosshair
//------------------------------------------------------------------------------
void RenderCrosshair (void)
{
	int x, y;
	
	// Get cursor position
	SDL_GetMouseState (&x, &y);
	
	SDL_Rect dstRect = {x - 10, y - 10, 0, 0};
	
	SDL_BlitSurface (pBitmaps, &CrosshairRect, pScreen, &dstRect);
}
//==============================================================================

//==============================================================================
// Blit blood decal to background
//------------------------------------------------------------------------------
void BloodDecal (Uint16 x, Uint16 y)
{
	SDL_Rect dstRect = {x, y + 5, 0, 0};
	SDL_BlitSurface (pBitmaps, &BloodDecalRect, pBackground, &dstRect);
}
//==============================================================================

//==============================================================================
// Blit explosion decal to background
//------------------------------------------------------------------------------
void ExplosionDecal (Uint16 x, Uint16 y)
{
	SDL_Rect dstRect = {x + 5, y + 10, 0, 0};
	SDL_BlitSurface (pBitmaps, &ExplosionDecalRect, pBackground, &dstRect);
}
//==============================================================================

//==============================================================================
// Blit sand decal to background
//------------------------------------------------------------------------------
void SandDecal (Uint16 x, Uint16 y)
{
	SDL_Rect dstRect = {x - 2, y - 4, 0, 0};
	SDL_BlitSurface (pBitmaps, &SandDecalRect, pBackground, &dstRect);
}
//==============================================================================

#ifdef SOUND
//==============================================================================
// Load sound (WAV)
//------------------------------------------------------------------------------
Mix_Chunk *LoadSound (string filename)
{
	string completeFilename = DATA_DIR;
	completeFilename.append (filename);
	
	Mix_Chunk *pTemp = Mix_LoadWAV (completeFilename.c_str ());
	if ( pTemp == NULL )
	{
		pTemp = Mix_LoadWAV (filename.c_str ());
		if ( pTemp == NULL )
		{
			cerr << "Unable to load sound " << completeFilename.c_str () << " or " << filename.c_str () << endl;
			throw runtime_error ("Unable to load resources");
		}
	}
	return pTemp;
}
//==============================================================================
#endif

//==============================================================================
// Save highscores
//------------------------------------------------------------------------------
void LoadHighscores (void)
{
	string filename = SCORE_DIR;
	filename.append ("oilwar.scores");
	
	ifstream fin ("oilwar.scores");
	if ( !fin )
	{
		fin.open (filename.c_str ());
		if ( !fin )
		{
			cerr << "Unable to open highscore file for reading" << endl;
			return;
		}
	}
	
	// Load kills
	fin.read (&HighscoreKills, sizeof (Uint16));
	// Load score
	fin.read (&HighscoreScore, sizeof (Uint16));		
	
	if ( !fin )
		cerr << "Unable to load highscores\n";
	
	// Close the file
	fin.close ();
}	
//==============================================================================

//==============================================================================
// Save highscores
//------------------------------------------------------------------------------
void SaveHighscores (void)
{
	string filename = SCORE_DIR;
	filename.append ("oilwar.scores");
	
	ofstream fout ("oilwar.scores", ios::nocreate);
	if ( !fout )
	{
		fout.open (filename.c_str (), ios::nocreate);
		if ( !fout )
		{
			cerr << "Unable to open highscore file for writing" << endl;
			return;
		}
	}
	
	// Save kills
	fout.write (&Kills, sizeof (Uint16));
	// Save score
	fout.write (&Score, sizeof (Uint16));		
	
	if ( !fout )
		cerr << "Unable to save highscores\n";
	
	// Close the file
	fout.close ();
}
//==============================================================================

//==============================================================================
// EOF
//==============================================================================
